/*
 *  The parser to analyze collected data
 *
 *  Copyright (C) 2006  Christophe Devine
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <termios.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include <X11/Xlib.h>

#include "config.h"
#include "rc4.h"
#include "sha1.h"

struct packet
{
    struct packet *prev;
    struct packet *next;
    struct session *sn;
    unsigned char ip[4];
    struct timeval tv;
    int ws_row, ws_col;
    int uid, pid;
    int dir, len;
    unsigned char *data;
};

struct session
{
    struct session *prev;
    struct session *next;
    struct packet *pkt_1st;
    struct packet *pkt_end;
    unsigned char hash[20];
    int session_id;
    int nb_packets;
    int is_closed;
};

struct options
{
    int fd_log;
    int session_id;
    int display_mode;
};

int nb_read = 0;
int got_sigwinch, got_sigalrm;
unsigned char buffer[BUFSIZE];
char fchr[] = "~ +~*";

void sighandler( int signum )
{
    if( signum == SIGWINCH )
        got_sigwinch = 1;

    if( signum == SIGALRM )
        got_sigalrm = 1;
}

struct packet *pt_1st, *pt_end;
struct packet *pt_prv, *pt_cur;

struct session *sn_1st, *sn_end;
struct session *sn_prv, *sn_cur;

int read_all( int fd, int remaining )
{
    int ret;

    while( nb_read < remaining )
    {
        ret = read( fd, buffer + nb_read, remaining - nb_read );

        if( ret < 0 )
        {
            perror( "read" );
            exit( 1 );
        }

        if( ret == 0 )
            return( 0 );

        nb_read += ret;
    }

    nb_read = 0;
    return( 1 );
}

int read_n_decrypt( int fd )
{
    int len;
    sha1_context sha1;
    struct rc4_state rc4;
    unsigned char src_ip[4];
    unsigned char pktkey[32];
    unsigned char sha1sum[20];

    /* read the encrypted data */

    if( read_all( fd, 6 ) == 0 )
    {
        usleep( 10000 );
        return( 0 );
    }

    memcpy( src_ip, buffer, 4 );

    len = buffer[4] + ( ( buffer[5] ) << 8 );

    if( len < 54 || len > 4096 )
    {
        fprintf( stderr, "invalid packet length %d, corrupted file?\n", len );
        return( 2 );
    }

    while( read_all( fd, len ) == 0 )
        usleep( 10000 );

    /* try to verify the SHA-1 checksum */

    memcpy( pktkey     , secret, 16 );
    memcpy( pktkey + 16, buffer, 16 );

    sha1_starts( &sha1 );
    sha1_update( &sha1, pktkey, 32 );
    sha1_finish( &sha1, pktkey );

    rc4_setup( &rc4, pktkey, 20 );
    rc4_crypt( &rc4, buffer + 16, len - 16 );

    sha1_starts( &sha1 );
    sha1_update( &sha1, buffer + 16, len - 16 - 20 );
    sha1_finish( &sha1, sha1sum );

    if( memcmp( sha1sum, buffer + len - 20, 20 ) != 0 )
    {
        fprintf( stderr, "SHA-1 checksum verification failed\n" );
        return( 3 );
    }

    /* add the packet to the linked list */

    pt_prv = pt_cur;
    pt_cur = pt_end = (struct packet *) malloc(
               sizeof( struct packet ) );

    memset( pt_cur, 0, sizeof( struct packet ) );

    pt_cur->prev = pt_prv;
    pt_cur->next = NULL;

    if( pt_1st == NULL )
        pt_1st = pt_cur;

    if( pt_prv != NULL )
        pt_prv->next = pt_cur;

    memcpy( pt_cur->ip, src_ip, 4 );

    /* convert from network to host order */

    pt_cur->tv.tv_sec =  ( buffer[16]       ) |
                         ( buffer[17] <<  8 ) |
                         ( buffer[18] << 16 ) |
                         ( buffer[19] << 24 );

    pt_cur->tv.tv_usec = ( buffer[20]       ) |
                         ( buffer[21] <<  8 ) |
                         ( buffer[22] << 16 ) |
                         ( buffer[23] << 24 );

    pt_cur->uid = ( buffer[24] ) | ( buffer[25] << 8 );
    pt_cur->pid = ( buffer[26] ) | ( buffer[27] << 8 );

    pt_cur->ws_row = ( buffer[28] ) | ( buffer[29] << 8 );
    pt_cur->ws_col = ( buffer[30] ) | ( buffer[31] << 8 );

    pt_cur->dir = buffer[32];
    pt_cur->len = len - 53;

    pt_cur->data = (unsigned char *) malloc( pt_cur->len + 1 );
    memcpy( pt_cur->data, buffer + 33, pt_cur->len );
    pt_cur->data[pt_cur->len] = '\0';

    /* find the relevant session, using hash(ip,uid,pid) as the key */

    sha1_starts( &sha1 );
    sha1_update( &sha1, (void *) pt_cur->ip, 4 );
    sha1_update( &sha1, (void *) &pt_cur->uid, sizeof( int ) );
    sha1_update( &sha1, (void *) &pt_cur->pid, sizeof( int ) );
    sha1_finish( &sha1, sha1sum );

    sn_cur = sn_end;

    while( sn_cur != NULL )
    {
        if( memcmp( sha1sum, sn_cur->hash, 20 ) == 0 )
            break;

        sn_cur = sn_cur->prev;
    }

    if( sn_cur == NULL )
    {
        /* none found, create a new session */

        sn_cur = (struct session *) malloc(
                    sizeof( struct session ) );

        memset( sn_cur, 0, sizeof( struct session ) );

        sn_cur->prev = sn_end;
        sn_cur->next = NULL;

        sn_cur->pkt_1st = pt_cur;
        sn_cur->pkt_end = pt_cur;

        memcpy( sn_cur->hash, sha1sum, 20 );

        if( sn_1st == NULL )
            sn_1st = sn_cur;

        if( sn_prv != NULL )
        {
            sn_cur->session_id = sn_prv->session_id + 1;
            sn_prv->next = sn_cur;
        }
        else
            sn_cur->session_id = 1;

        sn_end = sn_prv = sn_cur;
    }
    else
        sn_cur->pkt_end = pt_cur;

    sn_cur->nb_packets++;

    if( pt_cur->ws_row == 0 &&
        pt_cur->ws_col == 0 )
        sn_cur->is_closed = 1;

    pt_cur->sn = sn_cur;

    return( 1 );
}

int main( int argc, char *argv[] )
{
    int i, j, n;
    int clear_clock;
    int fast_forward;
    int line_size;
    char line[8192];
    char buf[8192], *s;
    struct options opt;
    struct winsize ws;
    struct timeval tv;
    struct termios tattr;
    struct termios tbttr;
    struct tm *lt;
    fd_set rd;

    Display *dpi;
    Window wnd;
    XWindowAttributes xwa;
    int nx, ny, bx, by;

    /* parse the program arguments */

    if( argc < 2 )
    {
        fprintf( stderr, "usage: %s <logfile> [SID] [mode]\n\n", argv[0] );
        fprintf( stderr, "mode can be 0 (listing) or 1 (interactive)\n" );
        return( 1 );
    }

    memset( &opt, 0, sizeof( opt ) );

    if( ( opt.fd_log = open( argv[1], O_RDONLY ) ) < 0 )
    {
        perror( "open" );
        return( 1 );
    }

    if( argc >= 3 ) opt.session_id   = atoi( argv[2] );
    if( argc >= 4 ) opt.display_mode = atoi( argv[3] );

    pt_1st = pt_prv = pt_cur = pt_end = NULL;
    sn_1st = sn_prv = sn_cur = sn_end = NULL;

    /* parse the logfile */

    while( read_n_decrypt( opt.fd_log ) > 0 );

    if( opt.session_id != 0 )
        goto have_sid;

    /* ask the user for the session id */

    fprintf( stderr, "\n SID        SOURCE IP    UID    PID    "
                     "START DATE      END DATE      DURATION\n\n" );

    sn_cur = sn_1st;

    while( sn_cur != NULL )
    {
        int nb_h, nb_m, nb_s;

        pt_prv = sn_cur->pkt_1st;
        pt_cur = sn_cur->pkt_end;

        n = pt_cur->tv.tv_sec - pt_prv->tv.tv_sec;

        nb_h = ( n / 3600 );
        nb_m = ( n - ( nb_h * 3600 ) ) / 60;
        nb_s = ( n - ( nb_h * 3600 ) - ( nb_m * 60 ) );

        sprintf( buf, "[%d.%d.%d.%d]",
                 pt_cur->ip[0], pt_cur->ip[1],
                 pt_cur->ip[2], pt_cur->ip[3] );

        lt = localtime( &pt_prv->tv.tv_sec );

        fprintf( stderr, "%4d %17s %5d (%5d)  %02d/%02d %02d:%02d | ",
                sn_cur->session_id, buf, pt_cur->uid, pt_cur->pid,
                lt->tm_mday, lt->tm_mon + 1,
                lt->tm_hour, lt->tm_min );

        lt = localtime( &pt_cur->tv.tv_sec );

        fprintf( stderr, "%02d/%02d %02d:%02d  %c",
                lt->tm_mday, lt->tm_mon + 1,
                lt->tm_hour, lt->tm_min,
                ( sn_cur->is_closed == 1 ) ? ' ' : 'X' );

        if( nb_h > 0 )  printf( "%2dh ", nb_h );
        else            printf( "    " );

        if( nb_m > 0 )  printf( "%2dm ", nb_m );
        else            printf( "    " );

        printf( "%02ds\n", nb_s );

        sn_cur = sn_cur->next;
    }

    fprintf( stderr, "\nIn interactive mode, use Enter to fast forward" \
                     ", Space to pause and q to quit.\nNote that xterm" \
                     " is required for window resizing.\n" );

    fprintf( stderr, "\nSession ID -> ");
    fflush( stderr );
    scanf( "%d", &opt.session_id );

    fprintf( stderr, "\nInteractive mode (y/n) ? " );
    scanf( "%s", buf );

    if( buf[0] == 'y' || buf[0] == 'Y' )
        opt.display_mode = 1;

    fprintf( stderr, "\n" );

have_sid:

    /* find the session matching the id */

    sn_cur = sn_1st;

    while( sn_cur != NULL )
    {
        if( sn_cur->session_id == opt.session_id )
            break;

        sn_cur = sn_cur->next;
    }

    if( sn_cur == NULL )
    {
        fprintf( stderr, "Unknown session id, exiting.\n\n" );
        return( 1 );
    }

    if( opt.display_mode == 1 )
        goto display_interactive;

    /* display a listing of input commands */

    line_size = 0;
    pt_cur = pt_1st;

    while( 1 )
    {
        while( pt_cur == NULL )
            read_n_decrypt( opt.fd_log );

        /* skip any packet not from the chosen session */

        if( pt_cur->sn->session_id != opt.session_id )
        {
            pt_cur = pt_cur->next;
            continue;
        }

        /* check if we reached the end of the session */

        if( pt_cur->ws_row == 0 &&
            pt_cur->ws_col == 0 )
            break;

        /* skip output packets */

        if( pt_cur->dir != 0 )
        {
            pt_cur = pt_cur->next;
            continue;
        }

        /* reconstruct the cmdline, assuming it ends with \r */

        if( line_size + pt_cur->len >= 8192 )
            line_size = 0;

        memcpy( line + line_size,
                pt_cur->data, pt_cur->len );

        line_size += pt_cur->len;
        line[line_size] = '\0';

        s = strchr( (char *) line, '\r' );

        if( s != NULL )
        {
            /* remove specials chars & apply backspace */

            n = s - (char *) &line[0];

            for( i = j = 0; i < n; i++, j++ )
            {
                if( j >= (int) sizeof( buf ) - 1 )
                    break;

                if( line[i] == 0x7f && j > 0 )
                {
                    j -= 2;
                    continue;
                }

                if( line[i] < 32 )
                    buf[j] = 32;
                else
                    buf[j] = line[i];
            }

            buf[j] = '\0';

            /* display the line and remove the consumed input */

            lt = localtime( &pt_cur->tv.tv_sec );

            printf( "%02d/%02d %02d:%02d:%02d -> %s\n",
                    lt->tm_mday, lt->tm_mon + 1,
                    lt->tm_hour, lt->tm_min, lt->tm_sec, buf );

            memcpy( buf, line, line_size );
            line_size -= ( n + 1 );
            memset( line, 0, 8192 );
            memcpy( line, buf + n + 1, line_size );
        }

        pt_cur = pt_cur->next;
    }

    fprintf( stderr, "\nEnd of session.\n\n" );
    close( opt.fd_log );
    return( 0 );

display_interactive:

    /* get the current xterm window attributes */

    dpi = NULL;

    ioctl( 0, TIOCGWINSZ, &ws );

    if( ( s = getenv( "WINDOWID" ) ) != NULL )
    {
        if( ( dpi = XOpenDisplay( NULL ) ) == NULL )
        {
            fprintf( stderr, "fatal: I have a WINDOWID but failed " \
                             "to open DISPLAY. Check xauth?\r\n" );
            return( 1 );
        }

        sscanf( s, "%lu", &wnd );

        XSetWindowBorderWidth( dpi, wnd, 0 );
        XSync( dpi, False );
        XGetWindowAttributes( dpi, wnd, &xwa );

        nx = xwa.width  / ws.ws_col;
        ny = xwa.height / ws.ws_row;

        bx = xwa.width  - nx * ws.ws_col;
        by = xwa.height - ny * ws.ws_row;
    }

    /* set the current tty in RAW mode */

    if( tcgetattr( 0, &tattr ) == 0 )
    {
        memcpy( &tbttr, &tattr, sizeof( tattr ) );

        tbttr.c_iflag |= IGNPAR;
        tbttr.c_iflag &= ~(ISTRIP|INLCR|IGNCR|ICRNL|IXON|IXANY|IXOFF);
        tbttr.c_lflag &= ~(ISIG|ICANON|ECHO|ECHOE|ECHOK|ECHONL|IEXTEN);
        tbttr.c_oflag &= ~OPOST;

        tbttr.c_cc[VMIN]  = 1;
        tbttr.c_cc[VTIME] = 0;

        tcsetattr( 0, TCSANOW, &tbttr );
    }

    fprintf( stderr, "\33c\33[1;1H\33[J\n" );

    /* display the full user session, takedown-like */

    pt_prv = NULL;
    pt_cur = pt_1st;

    fast_forward = 1;
    clear_clock  = 0;

    while( 1 )
    {
        /* see if data is available in stdin */

        FD_ZERO( &rd );
        FD_SET( 0, &rd );

        tv.tv_sec  = 0;
        tv.tv_usec = 0;

        if( select( 1, &rd, NULL, NULL, &tv ) > 0 )
        {
            read( 0, buf, 1 );

            if( buf[0] == '\3' || ( buf[0] | 0x20 ) == 'q' )
                break;

            if( buf[0] == '\r' )
            {
                /* multiply fast forward by 2 */

                fast_forward <<= 1;

                if( fast_forward == 8 )
                    fast_forward  = 1;
            }

            if( buf[0] == ' ' )
            {
                /* pause until space is pressed again */

                printf( "%c7\33[1;%dH\33[7m [PAUSED] \33[0m%c8",
                        0x1b, ws.ws_col - 9, 0x1b );
                fflush( stdout );

                while( 1 )
                {
                    if( read( 0, buf, 1 ) == 1 &&
                        buf[0] == ' ' )
                        break;
                    else
                        usleep( 10000 );
                }

                printf( "%c7\33[1;%dH\33[0m          %c8",
                        0x1b, ws.ws_col - 9, 0x1b );
                fflush( stdout );
            }
        }

        /* if EOF was reached, try to grab another packet */

        if( pt_cur == NULL )
        {
            FD_ZERO( &rd );
            FD_SET( opt.fd_log, &rd );

            tv.tv_sec  = 0;
            tv.tv_usec = 10000;

            if( select( opt.fd_log + 1, &rd, NULL, NULL, &tv ) <= 0 )
                continue;

            if( read_n_decrypt( opt.fd_log ) != 1 )
                continue;
        }

        /* skip if not from current session id */

        if( pt_cur != NULL )
        {
            if( pt_cur->sn->session_id != opt.session_id )
            {
                pt_cur = pt_cur->next;
                continue;
            }
        }

        /* test if end of session was reached */

        if( pt_cur->ws_row == 0 &&
            pt_cur->ws_col == 0 )
        {
            clear_clock = 1;
            printf( "%c7\33[1;%dH\33[7m = STOP = %c8\33[0m",
                    0x1b, ws.ws_col - 9, 0x1b );
            fflush( stdout );
        }

        /* skip if not an output (pty -> console) packet */

        if( pt_cur->dir != 1 )
        {
            pt_cur = pt_cur->next;
            continue;
        }

        if( pt_prv != NULL )
        {
            /* sleep at most 2 seconds */

            double tt = (double) ( pt_cur->tv.tv_sec  -
                                   pt_prv->tv.tv_sec  ) * 1000000.0f
                               + ( pt_cur->tv.tv_usec -
                                   pt_prv->tv.tv_usec );

            if( tt > 2000000.0f )
                tt = 2000000.0f;

            if( tt > 4000.0f )
            {
                tt = tt / (double) fast_forward;

                /* only sleep if tt > time granularity */

                tv.tv_sec  = tt / 1000000.0f;
                tv.tv_usec = tt - ( (double) tv.tv_sec * 1000000.0f );

                FD_ZERO( &rd );
                select( 0, &rd, NULL, NULL, &tv );
            }
        }

        if( clear_clock == 1 )
        {
            clear_clock = 0;
            printf( "%c7\33[1;%dH\33[0m          %c8",
                    0x1b, ws.ws_col - 9, 0x1b );
            fflush( stdout );
        }

        if( ws.ws_row != pt_cur->ws_row ||
            ws.ws_col != pt_cur->ws_col )
        {
            /* need to resize the current window */

            if( dpi == NULL )
            {
                fprintf( stderr, "fatal: resizing needed (%d x %d) but got no " \
                                 "DISPLAY.\r\nIs the console an xterm?\r\n", pt_cur->ws_col, pt_cur->ws_row );
                break;
            }

            got_sigwinch = 0;
            got_sigalrm  = 0;

            signal( SIGWINCH, sighandler );
            signal( SIGALRM,  sighandler );

            ws.ws_row = pt_cur->ws_row;
            ws.ws_col = pt_cur->ws_col;

            XMoveResizeWindow( dpi, wnd, 8, 8, nx * ws.ws_col + bx,
                                               ny * ws.ws_row + by );
            XSync( dpi, False );

            /* wait until xterm has finished resizing */

            alarm( 2 );
            while( got_sigwinch == 0 )
            {
                if( got_sigalrm == 1 )
                {
                    fprintf( stderr, "\r\nfatal: window resizing failed; " \
                                     "make sure xterm is not maximized.\r\n" );
                    goto exit_display;
                }
                usleep( 10000 );
            }
            alarm( 0 );

            ioctl( 0, TIOCGWINSZ, &ws );

            if( ws.ws_row != pt_cur->ws_row ||
                ws.ws_col != pt_cur->ws_col )
            {
                fprintf( stderr, "\r\nfatal: window resizing failed\r\n" );
                goto exit_display;
            }
        }

        write( 1, pt_cur->data, pt_cur->len );

        /* refresh the clock */

        if( strchr( (char *) pt_cur->data, 0x1b ) == NULL )
        {
            lt = localtime( &pt_cur->tv.tv_sec );

            clear_clock = 1;
            printf( "%c7\33[1;%dH\33[7m%c%02d:%02d:%02d%c\33[0m%c8",
                    0x1b, ws.ws_col - 9, fchr[fast_forward],
                    lt->tm_hour, lt->tm_min, lt->tm_sec,
                    fchr[fast_forward], 0x1b );
            fflush( stdout );
        }

        pt_prv = pt_cur;
        pt_cur = pt_cur->next;
    }

exit_display:

    /* restore the console attributes and exit */

    tcsetattr( 0, TCSANOW, &tattr );

    write( 1, "\33[0m\n", 5 );

    if( dpi != NULL )
        XCloseDisplay( dpi );

    close( opt.fd_log );

    return( 0 );
}
