/*
 *  The sh2logd PTY logging daemon
 *
 *  Copyright (C) 2006  Christophe Devine
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>

#include "config.h"
#include "rc4.h"
#include "sha1.h"

/*
 * Packet written to log:
 *
 *   0   -  3        Source IP address
 *   4   -  5        Total packet length
 *  -----------------------------------------}
 *   6   - 21        RC4 initialization vector
 *  -----------------------------------------}
 *  22   - 29        gettimeofday()          |
 *  30   - 31        real user id            |
 *  32   - 33        shell process id        |
 *  34   - 37        current window size     |  RC4-encrypted data
 *  38   - 38        direction (0=in,1=out)  |
 *  39   - 39+n-1    message contents        |
 *  39+n - 39+n+20   SHA-1 control sum       |
 *  -----------------------------------------}
 *
 *  The RC4 IV and the secret key are hashed together with SHA-1
 *  to produce a per-packet RC4 encryption/decryption key.
 */

void create_log( int *fd_log )
{
    char logfile[64];
    struct timeval tv;
    struct tm *lt;

    gettimeofday( &tv, NULL );
    lt = localtime( &tv.tv_sec );

    sprintf( logfile, "sh2log-%04d%02d%02d-%02d%02d%02d.bin",
             1900 + lt->tm_year, 1 + lt->tm_mon,
             lt->tm_mday, lt->tm_hour, lt->tm_min, lt->tm_sec );

    *fd_log = open( logfile, O_WRONLY | O_CREAT | O_APPEND,
                             S_IRUSR | S_IWUSR );

    if( *fd_log < 0 )
    {
        perror( "open" );
        exit( 1 );
    }
}

int main( void )
{
    socklen_t fromlen;
    int len, n, server, fd_log;
    unsigned long int nb_written;
    struct sockaddr_in client_addr;
    struct sockaddr_in server_addr;
    unsigned char buffer[BUFSIZE];
    unsigned char logbuf[BUFSIZE + 64];
    unsigned char pktkey[32];
    unsigned char sha1sum[40];
    struct rc4_state rc4;
    sha1_context sha1;
    struct timeval tv;

    create_log( &fd_log );
    nb_written = 0;

    /* start the UDP listening server */

    if( ( server = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP ) ) < 0 )
    {
        perror( "socket" );
        return( 1 );
    }

    n = 1;

    if( setsockopt( server, SOL_SOCKET, SO_REUSEADDR,
                    (void *) &n, sizeof( n ) ) < 0 )
    {
        perror( "setsockopt" );
        return( 1 );
    }

    server_addr.sin_family      = AF_INET;
    server_addr.sin_port        = htons( SERVER_PORT );
    server_addr.sin_addr.s_addr = htonl( INADDR_ANY );

    if( bind( server, (struct sockaddr *) &server_addr,
              sizeof( server_addr ) ) < 0 )
    {
        perror( "bind" );
        return( 1 );
    }

    if( fork() != 0 )
        exit( 0 );

    setsid();

    while( 1 )
    {
        fromlen = sizeof( client_addr );

        if( ( len = recvfrom( server, buffer, BUFSIZE, 0,
                              (struct sockaddr *) &client_addr,
                              &fromlen ) ) < 54 )
        {
            sleep( 1 );
            continue;
        }

        /* decrypt the buffer and verify the SHA-1 checksum */

        memcpy( pktkey,      secret, 16 );
        memcpy( pktkey + 16, buffer, 16 ); 

        sha1_starts( &sha1 );
        sha1_update( &sha1, pktkey, 32 );
        sha1_finish( &sha1, pktkey );

        rc4_setup( &rc4, pktkey, 20 );
        rc4_crypt( &rc4, buffer + 16, len - 16 );

        sha1_starts( &sha1 );
        sha1_update( &sha1, buffer + 16, len - 16 - 20 );
        sha1_finish( &sha1, sha1sum );

        if( memcmp( buffer + len - 20, sha1sum, 20 ) != 0 )
            continue;

        /* set the packet timestamp */

        gettimeofday( &tv, NULL );

        buffer[16] = ( tv.tv_sec        ) & 0xFF;   
        buffer[17] = ( tv.tv_sec  >>  8 ) & 0xFF;   
        buffer[18] = ( tv.tv_sec  >> 16 ) & 0xFF;   
        buffer[19] = ( tv.tv_sec  >> 24 ) & 0xFF;   

        buffer[20] = ( tv.tv_usec       ) & 0xFF;   
        buffer[21] = ( tv.tv_usec >>  8 ) & 0xFF;   
        buffer[22] = ( tv.tv_usec >> 16 ) & 0xFF;   
        buffer[23] = ( tv.tv_usec >> 24 ) & 0xFF;   

        /* re-encrypt the buffer, write IP + len + buffer */

        sha1_starts( &sha1 );
        sha1_update( &sha1, buffer + 16,  len - 16 - 20 );
        sha1_finish( &sha1, buffer + len - 20 );

        rc4_setup( &rc4, pktkey, 20 );
        rc4_crypt( &rc4, buffer + 16, len - 16 );

        memcpy( logbuf, &client_addr.sin_addr.s_addr, 4 );

        logbuf[4] = ( len      ) & 0xFF;
        logbuf[5] = ( len >> 8 ) & 0xFF;

        memcpy( logbuf + 6, buffer, len );
        write( fd_log, logbuf, 6 + len );

        nb_written += ( 6 + len );

        if( nb_written > MAX_LOG_SIZE )
        {
            nb_written = 0;
            close( fd_log );
            create_log( &fd_log );
        }
    }

    return( 0 );
}
